# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: agg [-g GROUPING_FUNCTION] INITIAL_VALUE AGGREGATION_FUNCTION
       agg [-c GROUPING_FUNCTION] INITIAL_VALUE AGGREGATION_FUNCTION
    
Aggregates objects from stdin. If GROUPING_FUNCTION is omitted, then
one output object is generated by initializing an accumulator to INITIAL_VALUE
and then combining the accumulator with stdin objects using AGGREGATION_FUNCTION.
AGGREGATION_FUNCTION takes two inputs, the current value of the accumulator and
an object from stdin.

Example: If the stdin objects are integers 1, 2, 3, then the sum of the integers
is computed as follows:

... ^ agg 0 'sum, x: sum + x'

which yields 6.

If GROUPING_FUNCTION is specified, then a set of accumulators is maintained,
one for each value of GROUPING_FUNCTION. Each output object is a tuple with
two parts, the group value and the accumulated value for the group.

Example: If the stdin objects are ('a', 1), ('a', 2), ('b', 3), ('b', 4), then
the sum of ints for each string is computed as follows:

... ^ agg -g 't: t[0]' 0 'sum, t: sum + t[1]'

which yields ('a', 3), ('b', 7).
 If the grouping function is specified with the -g flag, then agg generates its
output when the input stream has ended. (It has to, because group members map
appear in any order.) In some situations however, group members appear consecutively,
and it is useful to get output earlier. If group members are known to be consecutive,
then the group function can be specified using the -c flag.
"""

import osh.oshargs
import osh.oshobjects

_wrap_if_necessary = osh.oshobjects.wrap_if_necessary

def agg(*args):
    return _Agg().process_args(*args)

def group(function_spec):
    return osh.oshargs.Option('-g', function_spec)

class _Agg(osh.oshobjects.Op):

    _aggregate = None


    # object interface

    def __init__(self):
        osh.oshobjects.Op.__init__(self, 'g:c:', (2, 2))


    # BaseOp interface

    def doc(self):
        return __doc__

    def setup(self):
        args = self.args()
        grouping_function = args.function_arg('-g')
        consecutive_grouping_function = args.function_arg('-c')
        initial_value = _wrap_if_necessary(args.next_eval())
        aggregation_function = args.next_function()
        if grouping_function and consecutive_grouping_function:
            self.usage()
        if initial_value is None or aggregation_function is None:
            self.usage()
        if args.has_next():
            self.usage()
        if grouping_function:
            self._aggregate = _GroupingAggregate(
                self,
                grouping_function,
                initial_value,
                aggregation_function)
        elif consecutive_grouping_function:
            self._aggregate = _ConsecutiveGroupingAggregate(
                self,
                consecutive_grouping_function,
                initial_value,
                aggregation_function)
        else:
            self._aggregate = _SimpleAggregate(
                self,
                initial_value,
                aggregation_function)


    def receive(self, object):
        self._aggregate.receive(object)

    def receive_complete(self):
        self._aggregate.receive_complete()


class _GroupingAggregate:
    _command = None
    _groupFunction = None
    _initial_value = None
    _aggregate_function = None
    _sum = None

    def __init__(self, command, groupFunction, initial_value, aggregate_function):
        self._command = command
        self._groupFunction = groupFunction
        self._initial_value = initial_value
        self._aggregate_function = aggregate_function
        self._sum = {}

    def receive(self, object):
        group = self._groupFunction(object)
        sum = self._sum.get(group, self._initial_value)
        self._sum[group] = _wrap_if_necessary(self._aggregate_function(tuple(sum) + tuple(object)))

    def receive_complete(self):
        for group, sum in self._sum.iteritems():
            self._command.send(_wrap_if_necessary(group) + tuple(sum))

class _ConsecutiveGroupingAggregate:
    _command = None
    _groupFunction = None
    _initial_value = None
    _aggregate_function = None
    _group = None
    _sum = None

    def __init__(self, command, groupFunction, initial_value, aggregate_function):
        self._command = command
        self._groupFunction = groupFunction
        self._initial_value = initial_value
        self._aggregate_function = aggregate_function
        self._group = None
        self._sum = None

    def receive(self, object):
        newGroup = self._groupFunction(object)
        if self._group is None or self._group != newGroup:
            if self._group is not None:
                self._command.send(_wrap_if_necessary(self._group) + tuple(self._sum))
            self._group = newGroup
            self._sum = self._initial_value
        self._sum = _wrap_if_necessary(self._aggregate_function(tuple(self._sum) + tuple(object)))

    def receive_complete(self):
        if self._group is not None:
            self._command.send(_wrap_if_necessary(self._group) + tuple(self._sum))

class _SimpleAggregate:
    _command = None
    _initial_value = None
    _aggregate_function = None
    _sum = None

    def __init__(self, command, initial_value, aggregate_function):
        self._command = command
        self._initial_value = initial_value
        self._aggregate_function = aggregate_function
        self._sum = initial_value

    def receive(self, object):
        self._sum = _wrap_if_necessary(self._aggregate_function(tuple(self._sum) + tuple(object)))

    def receive_complete(self):
        self._command.send(self._sum)
