# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: installosh  -c CLUSTER [-u USER] [OSH_CONFIG_FILE]
"""

import popen2
import sys
import threading

import osh.oshobjects
import progtrack

DISTRIBUTION_FILENAME = 'osh-' + osh.oshobjects.OSH_VERSION + '.tar.gz'
DEFAULT_CONFIG_FILE = '~/.oshrc'
PACKAGE_HOME = '/usr/share/osh'

def _ssh(host, command):
    ssh_command = 'ssh %s -l root %s ' % (host, command)
    read, write, err = popen2.popen3(ssh_command)
    err_lines = err.readlines()
    read.close()
    write.close()
    err.close()
    if len(err_lines) > 0:
        raise Exception(' '.join(err_lines))

def _scp(user, host, source, target):
    scp_command = 'scp %s %s@%s:%s' % (source, user, host, target)
    read, write, err = popen2.popen3(scp_command)
    err_lines = err.readlines()
    read.close()
    write.close()
    err.close()
    if len(err_lines) > 0:
        raise Exception(' '.join(err_lines))

def installosh():
    return _InstallOsh()

class _InstallOsh(osh.oshobjects.RemoteOp):

    _config_file = None
    _host_counter = None
    _ui = None

    # object interface

    def __init__(self):
        osh.oshobjects.RemoteOp.__init__(self, 'c:u:', (0, 1))
    

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.oshobjects.RemoteOp.setup(self)
        if self.args().has_next():
            self._config_file = self.args().next_string()
        else:
            self._config_file = DEFAULT_CONFIG_FILE
        self._host_counter = 0
        self._ui = progtrack.ProgressTrackingUI('installosh')
        self._ui.add_column('host', 25)
        self._ui.add_column(['create', 'osh dir'], 10)
        self._ui.add_column(['copy', 'package'], 10)
        self._ui.add_column(['copy', 'installer'], 10)
        self._ui.add_column(['run', 'installer'], 10)
        self._ui.add_column(['copy', 'conf file'], 10)
        for host in self.hosts():
            self._ui.add_row(host.name())

    # generator interface

    def execute(self):
        self._ui.start()
        try:
            osh.oshobjects.RemoteOp.execute(self)
        finally:
            self._ui.stop()


    # RemoteOp interface

    def action(self, host):
        stage = 0
        try:
            # Remove existing /usr/share/osh directory
            stage += 1
            _ssh(host.address(),
                 '"rm -rf %s"' % PACKAGE_HOME)
            # Create directory /usr/share/osh
            _ssh(host.address(),
                 '"mkdir %s"' % PACKAGE_HOME)
            self._ui.ok(host.name(), stage)
            # Copy package to /usr/share/osh
            stage += 1
            _scp(self.user(),
                 host.address(),
                 '%s/%s' % (PACKAGE_HOME, DISTRIBUTION_FILENAME),
                 PACKAGE_HOME)
            self._ui.ok(host.name(), stage)
            # Copy remoteinstall script to /usr/share/osh
            stage += 1
            _scp(self.user(),
                 host.address(),
                 '%s/remoteinstall' % PACKAGE_HOME,
                 PACKAGE_HOME)
            self._ui.ok(host.name(), stage)
            # Run remoteinstall script
            stage += 1
            _ssh(host.address(),
                 '"pushd %s; ./remoteinstall; popd"' % PACKAGE_HOME)
            self._ui.ok(host.name(), stage)
            # Copy config file to home directory
            stage += 1
            _scp(self.user(),
                 host.address(),
                 self._config_file,
                 '~')
            self._ui.ok(host.name(), stage)
        except:
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            message = str(exc_value).strip()
            self._ui.error(host.name(), stage, message)
