# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import curses
import sys
import threading

EXIT_MESSAGE = 'Press any key to exit'

lock = threading.Condition(threading.RLock())

class _Column:
    _label = None
    _width = None
    _offset = None

    def __init__(self, label, width):
        if type(label) == str:
            label = [label]
        self._label = label
        self._width = width

    def label(self):
        return self._label

    def width(self):
        return self._width

    def offset(self):
        return self._offset

    def set_offset(self, offset):
        self._offset = offset

class _Row:
    _label = None
    _position = None

    def __init__(self, label, position):
        self._label = label
        self._position = position

    def label(self):
        return self._label

    def position(self):
        return self._position

class ProgressTrackingUI:
    _started = False
    _screen = None
    _window = None
    _host_counter = None
    _columns = None
    _rows = None
    _messages = None
    _max_label_height = 0
    _footnote_count = 0

    # object interface

    def __init__(self, title):
        self._host_counter = 0
        self._columns = []
        self._rows = {}
        self._messages = []
        self.start_curses()

    # progress_tracking_ui interface
    
    def start(self):
        self.draw(self._max_label_height, 0, '+')
        self.draw(self.footnote_base() - 1, 0, '+')
        offset = 0
        for column in self._columns:
            label = column.label()
            width = column.width()
            column.set_offset(offset + 1)
            i = 0
            for line in label:
                self.draw(self._max_label_height - len(label) + i,
                          offset + 1,
                          label[i])
                i += 1
            self.draw(self._max_label_height, offset + 1, ('-' * width ) + '+')
            self.draw(self.footnote_base() - 1, offset + 1, ('-' * width ) + '+')
            offset += width + 1
        i = 0
        row_label_width = self._columns[0].width()
        for label, row in self._rows.items():
            line = self.row_base() + i
            self.draw(line, 1, label[:row_label_width])
            self.draw(line, 0, '|')
            for column in self._columns:
                self.draw(line, column.offset() + column.width(), '|')
            i += 1
        self._started = True

    def stop(self):
        self.draw(self.first_blank_line(), 0, EXIT_MESSAGE)
        self._window.getch(self.footnote_base() + self._footnote_count,
                           len(EXIT_MESSAGE) + 1)
        self.stop_curses()
        
    def add_column(self, label, width):
        # Single-line label: string
        # Multi-line label: sequence of strings
        self._columns.append(_Column(label, width))
        height = 1
        if type(label) != str:
            height = len(label)
        self._max_label_height = max(self._max_label_height, height)

    def add_row(self, label):
        self._rows[label] = _Row(label, len(self._rows))

    def ok(self, row_label, column_number):
        self.check_started()
        lock.acquire()
        row = self._rows[row_label]
        column = self._columns[column_number]
        self.draw(self.row_base() + row.position(),
                  column.offset(),
                  '*' * column.width())
        lock.release()

    def error(self, row_label, column_number, message):
        self.check_started()
        lock.acquire()
        row = self._rows[row_label]
        column = self._columns[column_number]
        footnote_id = self._footnote_count + 1
        self.draw(self.row_base() + row.position(),
                  column.offset(),
                  str(footnote_id))
        self.draw(self.footnote_base() + self._footnote_count,
                  0,
                  '%s) %s: %s' % (self._footnote_count + 1, row_label, message))
        self._footnote_count = footnote_id
        lock.release()

    # For use by this class

    def start_curses(self):
        try:
            self._screen = curses.initscr()
            curses.noecho()
            curses.cbreak()
            self._screen.keypad(1)
            self._window = curses.newwin(0, 0)
        except Exception, e:
            self.stop_curses()
            traceback.print_exc(file = sys.stderr)

    def stop_curses(self):
        self._screen.keypad(0)
        curses.echo()
        curses.nocbreak()
        curses.endwin()

    def check_started(self):
        if not self._started:
            self.stop_curses()
            raise 'UI not started yet'

    def row_base(self):
        return self._max_label_height + 1

    def footnote_base(self):
        return self.row_base() + len(self._rows) + 1

    def first_blank_line(self):
        return self.footnote_base() + self._footnote_count

    def draw(self, line, position, message):
        self._window.addstr(line, position, message)
        self._window.refresh()

# Example
# ui = progress_tracking_ui('test')
# ui.add_column('row', 10)
# ui.add_column(['foo', 'bar'], 5)
# ui.add_column(['three', 'line', 'header'], 8)
# ui.add_row('aa')
# ui.add_row('bb')
# ui.add_row('cc')
# ui.start()
# ui.ok('aa', 1)
# ui.ok('bb', 2)
# ui.ok('bb', 1)
# ui.ok('cc', 1)
# ui.error('cc', 2, 'oops')
# ui.error('aa', 2, 'aah!')
# ui.stop()
