# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: squish [SQUISH_SPEC]

squish receives a stream of sequences through stdin. Each sequence
is reduced to a single value through the use of the Python
function reduce. The SQUISH_SPEC describes the reduction to be done.

For example, if one of the input objects is [1, 2, 3, 4], then

    squish '+'

will generate 10 (1 + 2 + 3 + 4), while

    squish '*'

will generate 24 (1 * 2 * 3 * 4).

The result is exactly equivalent to what would be produced by using map, e.g.

    map 'list: squish(lambda a, b: a + b, list)'

but this is clearly involves more typing.

If the input sequence contains sequences, then the SQUISH_SPEC can specify
multiple operations to do multiple reductions at once. For example, if
one input sequence is [[10, 20, 30], [1, 100, 1000], [111, 222, 333]]
then

    squish '+ max min'

will produce [122, 222, 30]. 122 is 10 + 1 + 111, 222 is max(20, 100, 222), and
30 is min(30, 1000, 333).

The functions which can be used in a SQUISH_SPEC are any operator built in
to Python (e.g. +, *, -, and so on, although clearly some operators,
such as -, should be avoided); as well as max and min.

If SQUISH_SPEC is omitted, then a default value of + is used.
"""

import osh.oshobjects

# %s is an expression in terms of x and y, e.g. x + y.
SQUISH_FUNCTION_TEMPLATE = 'lambda input: reduce(lambda x, y: (%s), input)'

def squish(*args):
    return _Squish().process_args(*args)

class _Squish(osh.oshobjects.Op):

    _squisher = None


    # object interface

    def __init__(self):
        osh.oshobjects.Op.__init__(self, '', (0, 1))


    # OshCommand interface
    
    def doc(self):
        return __doc__

    def setup(self):
        args = self.args()
        if args.has_next():
            squish_spec = args.next_string()
            if args.has_next():
                self.usage()
        else:
            squish_spec = '+'
        squish_ops = [s.strip() for s in squish_spec.split()]
        if len(squish_ops) == 1:
            self._squisher = _object_squisher(squish_ops[0])
        else:
            self._squisher = _sequence_squisher(squish_ops)


    # Receiver interface
    
    def receive(self, object):
        self.send(self._squisher(object))

def _squish_expression(op, x, y):
    if op == 'max' or op == 'min':
        expr = '%s(%s, %s)' % (op, x, y)
    else:
        expr = '%s %s %s' % (x, op, y)
    return expr
        

def _object_squisher(op):
    squish_function = SQUISH_FUNCTION_TEMPLATE % _squish_expression(op, 'x', 'y')
    return eval(squish_function)

def _sequence_squisher(ops):
    squish_expressions = []
    position = 0
    for op in ops:
        squish_expressions.append(
            _squish_expression(op, ('x[%d]' % position), ('y[%d]' % position)))
        position += 1
    squish_function = SQUISH_FUNCTION_TEMPLATE % ', '.join(squish_expressions)
    return eval(squish_function)

